use anyhow::Result;
use utilities::prelude::*;

use super::{Music, Sound};

use std::collections::HashMap;
use std::sync::Arc;

#[derive(Copy, Clone, Debug)]
pub enum SoundInterpretation {
    Generic,
    Spatial,
}

#[derive(Debug)]
// pub(crate) struct VolumeHandler {
pub struct VolumeHandler {
    sound_volumes: HashMap<String, f32>,
}

impl VolumeHandler {
    pub(crate) fn new(volume_info: HashMap<String, f32>) -> Self {
        VolumeHandler {
            sound_volumes: volume_info,
        }
    }

    pub(crate) fn set_volume(
        &mut self,
        sound_type: &str,
        volume: f32,
        audio_objects: &mut AudioObjects,
    ) -> Result<()> {
        // check map if sound type is already present
        let changed = match self.sound_volumes.get_mut(sound_type) {
            Some(stype) => {
                // set volume
                if *stype != volume {
                    *stype = volume;

                    true
                } else {
                    false
                }
            }
            None => {
                // insert volume for requested sound type
                self.sound_volumes.insert(sound_type.to_string(), volume);

                true
            }
        };

        if changed {
            if sound_type == "music" {
                audio_objects.apply_to_music(|music| music.set_volume(volume))?;
            } else {
                audio_objects.apply_to_sound(Some(sound_type), |sound| sound.set_volume(volume))?;
            }
        }

        Ok(())
    }

    pub(crate) fn volume(&self, sound_type: &str) -> f32 {
        self.sound_volumes[sound_type]
    }
}

pub trait ClearQueueAudioObject: Send + Sync {
    fn end_looping(&self) -> Result<()>;
    fn is_playing(&self) -> Result<bool>;
}

#[derive(Default)]
pub struct AudioObjects {
    // sound handling
    sounds: HashMap<String, Vec<Arc<Sound>>>,

    // music handling
    music: Vec<Arc<Music>>,

    clear_queue: Vec<Arc<dyn ClearQueueAudioObject>>,
}

impl AudioObjects {
    pub fn add_sound(&mut self, sound: Arc<Sound>) {
        match self.sounds.get_mut(sound.sound_type()) {
            Some(sounds) => sounds.push(sound),
            None => {
                self.sounds
                    .insert(sound.sound_type().to_string(), vec![sound]);
            }
        }
    }

    pub fn add_music(&mut self, music: Arc<Music>) {
        self.music.push(music);
    }

    pub fn apply_to_sound<F>(&self, sound_type: Option<&str>, f: F) -> Result<()>
    where
        F: Fn(&Arc<Sound>) -> Result<()>,
    {
        match sound_type {
            Some(sound_type) => {
                if let Some(sounds) = self.sounds.get(sound_type) {
                    for sound in sounds.iter() {
                        f(sound)?;
                    }
                }
            }
            None => {
                for sounds in self.sounds.values() {
                    for sound in sounds.iter() {
                        f(sound)?;
                    }
                }
            }
        }

        Ok(())
    }

    pub fn apply_to_music<F>(&self, f: F) -> Result<()>
    where
        F: Fn(&Arc<Music>) -> Result<()>,
    {
        for music in self.music.iter() {
            f(music)?;
        }

        Ok(())
    }

    pub fn clear(&mut self) {
        self.sounds.clear();
        self.music.clear();
    }

    pub fn remove_sound(&mut self, sound: &Arc<Sound>) -> Result<()> {
        if let Some(sounds) = self.sounds.get_mut(sound.sound_type()) {
            if let Some(old_sound) = erase_arc(sounds, sound) {
                old_sound.end_looping()?;
                self.clear_queue.push(old_sound);
            }
        }

        Ok(())
    }

    pub fn check_clear_queue(&mut self) -> Result<()> {
        let mut new_queue = Vec::new();

        for sound in self.clear_queue.iter() {
            if sound.is_playing()? {
                new_queue.push(sound.clone());
            }
        }

        self.clear_queue = new_queue;

        Ok(())
    }
}
